# -*- coding: utf-8 -*-
# distutils: language=c++
# cython: language_level=3, boundscheck=False
# cython: c_string_type=unicode, c_string_encoding=utf8
# Created by huangzhibo on 2022/01/01
"""
    Provides access to the cgef_adjust interface.
"""

from .cgef_adjust cimport *
import numpy as np
cimport numpy as np
from cython cimport view
from typing import List
from libcpp cimport bool
from gefpy.cgef_adjust_helper import FilterGeneData

cdef class CgefAdjust:
    cdef cellAdjust *c_instance 

    def __cinit__(self):
        self.c_instance = new cellAdjust()

    def __init__(self):
        pass

    def __dealloc__(self):
        del self.c_instance

    """
        1. cell adjust
    """
    def get_cell_data(self, bgef, cgef):
        """
        Get raw cell data from cgef and bgef file.

        :param bgef: the bgef file path
        :param cgef: the cgef file path

        :returns: (genelist, vec_cell)
        """

        self.c_instance.readBgef(bgef)
        self.c_instance.readCgef(cgef)
        cdef vector[cellgem_label] vec_cell
        cdef vector[string] genelist
        self.c_instance.getCellLabelgem(genelist, vec_cell)
        return np.asarray(genelist), np.asarray(vec_cell)

    def write_cgef_adjustdata(self, path, celldata, dnbdata, outline_path=''):
        """
        write the adjust cell data to cgef

        :param path: set the Output path
        :param celldata: input the cell data
        :param dandata: input the dandata
        """
        cdef Cell [:] cell = celldata
        cdef DnbExpression [:] dnb = dnbdata
        self.c_instance.writeCellAdjust(path, outline_path, &cell[0], cell.shape[0], &dnb[0], dnb.shape[0])

    """
        2. lasso generate gef
    """
    def set_lasso_binsize(self, binsize):
        cdef vector[int] vec
        for t in binsize:
            vec.push_back(int(t))
        self.c_instance.setLassoBinsize(vec)

    def create_Region_Bgef_legacy(self, inpath, outpath, pos):
        cdef vector[vector[int]] vec
        for t in pos:
            vec.push_back(t)

        self.c_instance.readBgef(inpath)
        self.c_instance.getRegionGenedata(vec)
        self.c_instance.createRegionGef(outpath)

    def generate_region_bgef(self,input_file:str,output_file:str,flat_polygons) -> bool:
        cdef vector[vector[int]] vec
        for t in flat_polygons:
            vec.push_back(t)
        cdef bool ret = self.c_instance.generate_bgef_with_lasso(input_file,flat_polygons,output_file)
        return ret

    def create_Region_Bgef(self,input_file:str,output_file:str,flat_polygons) -> bool:
      print("recommend to use function generate_region_cgef instead creeate_Region_Cgef")
      return self.generate_region_bgef(input_file,output_file,flat_polygons)

    def generate_region_cgef(self,input_file:str,output_file:str,lasso_flat_polygons:List[List[int]]) -> bool:
        cdef vector[vector[int]] _cpp_lasso_flat_polygons
        for flat_polygon in lasso_flat_polygons:
            _cpp_lasso_flat_polygons.push_back(flat_polygon)
        cdef bool ret
        ret  = self.c_instance.generate_cgef_with_lasso(input_file,_cpp_lasso_flat_polygons,output_file)
        return ret
            

    def create_Region_Cgef(self, inpath, outpath, pos) -> bool:
        # cdef vector[vector[int]] vec
        # for t in pos:
        #     vec.push_back(t)
        
        # self.c_instance.getRegionCelldata(vec)
        # self.c_instance.readRawCgef(inpath)
        # self.c_instance.writeToCgef(outpath)
        print("recommend to use function generate_region_cgef instead creeate_Region_Cgef")
        return self.generate_region_cgef(inpath,outpath,pos)

    """
        3. lasso get data
    """
    def get_regiondata_frombgef(self, inpath, bin, thcnt, pos, isIndex = False):
        cdef vector[vector[int]] vec
        for t in pos:
            vec.push_back(t)

        cdef vector[vector[int]] vecdata_idx
        cdef vector[sapBgefData] vecdata
        cdef float region_area = 0.0
        if isIndex:
            self.c_instance.getSapRegionIndex(inpath, bin, thcnt, vec, vecdata_idx)
            return np.asarray(vecdata_idx)
        else:
            self.c_instance.getSapRegion(inpath, bin, thcnt, vec, vecdata, region_area)
            return np.asarray(vecdata), region_area
    
    def get_regiondata_fromcgef(self, input_path, pos):
        cdef vector[vector[int]] vec
        for t in pos:
            vec.push_back(t)
        
        self.c_instance.readRawCgef(input_path)
        self.c_instance.getRegionCelldataSap(vec)
        cdef sapCgefData vecdata
        self.c_instance.getSapCellbinRegion(vecdata)
        return np.asarray(vecdata)

    """
        4. get lasso region gene info, sort by mid
    """
    def get_multilabel_regiondata_bgef(self, inpath, pos, bin=1, thcnt=4):
        cdef vector[vector[int]] vec
        for t in pos:
            vec.push_back(t)

        cdef vector[LabelGeneData] region_data
        cdef unsigned int total_mid = 0
        self.c_instance.getMultiLabelInfoFromBgef(inpath, vec, region_data, total_mid, bin, thcnt)
        return np.asarray(region_data), total_mid

    def get_multilabel_regiondata_cgef(self, inpath, pos):
        cdef vector[vector[int]] vec
        for t in pos:
            vec.push_back(t)
        
        # self.c_instance.readRawCgef(inpath)
        # self.c_instance.getRegionCelldataSap(vec)
        cdef vector[LabelCellData] vecdata
        cdef vector[LabelCellDataSum] total_data
        self.c_instance.getMultiLabelInfoFromCgef(inpath, vec, vecdata, total_data)
        return np.asarray(vecdata), np.asarray(total_data)

    """
        5. get cord by clusterid from cgef
    """
    def get_position_by_clusterid(
            self,
            input_file:str,
            selected_cluster_ids:List[int],
            cluster_id_source_path:str="obs/leiden/codes",
            coordinate_source_path:str="/obsm/spatial") -> None:
        cdef vector[int] vec
        for t in selected_cluster_ids:
            vec.push_back(t)

        cdef vector[vector[int]] region_data
        self.c_instance.GetPositionIndexByClusterId(input_file,cluster_id_source_path,coordinate_source_path, vec, region_data)
        return np.asarray(region_data)

    """
        6. generate gef by filter mid, contain process rate
    """
    def generate_filter_bgef_by_midcnt(self,inpath, outpath, binsize, filter_data, only_filter=False,filter_with_gene_name:bool=True):
        """
        Args:
            filter_with_gene_name:bool,if True,filter by name,else filter by gene_id
        """
        cdef vector[MidCntFilter] filter_genes
        cdef MidCntFilter tmp
        for t in filter_data:
            tmp.gene_name = str(t['Gene'])
            tmp.max_mid = int(t['MaxFilterMID'])
            tmp.min_mid = int(t['MinFilterMID'])
            filter_genes.push_back(tmp)
        ret = self.c_instance.GenerateFilterBgefFileByMidCount(inpath, outpath, binsize, filter_genes, only_filter,filter_with_gene_name)
        return ret
    
    def better_generate_filter_bgef_by_mid_count(self,
                                               input_file:str,
                                               output_file:str,
                                               bin_size:int,
                                               filter_genes:List[FilterGeneData],
                                               gene_batch_size:int=4096,
                                               gene_exp_batch_size:int=4096 * 10,
                                               gene_stat_batch_size:int=4096,
                                               block_size:int=2048,bin1_tolearnce=10,
                                               run_with_async=False) -> bool:
        cdef vector[MidCntFilter] c_filter_genes
        c_filter_genes.reserve(len(filter_genes))
        cdef MidCntFilter c_filter_gene
        for filter_gene in filter_genes:
            c_filter_gene.gene_name = filter_gene.gene_str
            c_filter_gene.min_mid = filter_gene.min_mid_count
            c_filter_gene.max_mid = filter_gene.max_mid_count
            c_filter_genes.push_back(c_filter_gene)
        self.c_instance.better_generate_filter_bgef_by_mid_count(
            input_file,output_file,bin_size,c_filter_genes,gene_batch_size,
            gene_exp_batch_size,gene_stat_batch_size,block_size,bin1_tolearnce,run_with_async)


    def get_run_status(self) -> int:
        return self.c_instance.get_run_status()

    def get_run_detail_str(self) -> str:
        return self.c_instance.get_run_detail_str()

    def get_filter_bgef_process_rate(self):
        ret = self.c_instance.GenerateFilterBgefDuration()
        return int(ret)

    """
        7. lasso generate gef, for stereomap offline, contain process rate
    """
    def generate_bgef_by_lasso(self, inpath, outpath, pos):
        cdef vector[vector[int]] vec
        for t in pos:
            vec.push_back(t)
        ret = self.c_instance.GenerateBgefByLasso(inpath, outpath, vec)
        return ret

    def get_lasso_bgef_process_rate(self):
        ret = self.c_instance.GenerateLassoBgefDuration()
        return int(ret)

    """
        8. for stereopy, generate gef by coordinate, more accurate than lasso
    """
    def generate_bgef_by_coordinate(self, inpath, outpath, cord, bin_size):
        cdef vector[vector[int]] vec
        for t in cord:
            vec.push_back(t)
        ret = self.c_instance.createRegionBgefByCord(inpath, outpath, vec, bin_size)
        return int(ret)

    def generate_cgef_by_coordinate(self, inpath, outpath, cord):
        cdef vector[vector[int]] vec
        for t in cord:
            vec.push_back(t)
        ret = self.c_instance.createRegionCgefByCord(inpath, outpath, vec)
        return int(ret)